@@ -11,7 +11,7 @@ Authorization Server / Resource Server with Mojolicious
 
 =head1 VERSION
 
-0.11
+0.20
 
 =head1 SYNOPSIS
 
@@ -79,7 +79,7 @@ Then in your controller:
 
 This plugin implements the OAuth2 "Authorization Code Grant" flow as described
 at L<http://tools.ietf.org/html/rfc6749#section-4.1>. It is not a complete
-implementation of RFC6749, as it is rather large in scope. However the extra
+implementation of RFC6749, as that is rather large in scope. However the extra
 functionality and flows may be added in the future.
 
 This plugin enables you to easily (?) write an OAuth2 Authorization Server (AS)
@@ -108,6 +108,14 @@ can also see the tests and examples included with this distribution. OAuth2
 seems needlessly complicated at first, hopefully this plugin will clarify the
 various steps and simplify the implementation.
 
+If you would still like to use the plugin in an easy way, but also have ACs and
+ATs persistent across restarts and shared between multi processes then you can
+supply a jwt_secret. What you lose when doing this is the ability for tokens to
+be revoked. You could implement the verify_auth_code and verify_access_token
+methods to handle the revoking in your app. So that would be halfway between
+the "simple" and the "realistic" way. L<CLIENT SECRET, TOKEN SECURITY, AND JWT>
+has more detail about JWTs.
+
 Note that OAuth2 requires https, so you need to have the optional Mojolicious
 dependency required to support it. Run the command below to check if
 L<IO::Socket::SSL> is installed.
@@ -120,6 +128,12 @@ The plugin takes several configuration options. To use the plugin in a realistic
 way you need to pass several callbacks, documented in L<REQUIRED FUNCTIONS>, and
 marked here with a *
 
+=head2 jwt_secret
+
+This is optional. If set JWTs will be returned for the auth codes, access, and
+refresh tokens. JWTs allow you to validate tokens without doing a db lookup, but
+there are certain considerations (see L<CLIENT SECRET, TOKEN SECURITY, AND JWT>)
+
 =head2 authorize_route
 
 The route that the Client calls to get an authorization code. Defaults to
@@ -197,17 +211,20 @@ use warnings;
 use base qw/ Mojolicious::Plugin /;
 
 use Mojo::URL;
+use Mojo::JWT;
 use Time::HiRes qw/ gettimeofday /;
 use MIME::Base64 qw/ encode_base64 decode_base64 /;
 use Carp qw/ croak /;
 use Crypt::PRNG qw/ random_string /;
+use Try::Tiny;
 
-our $VERSION = '0.11';
+our $VERSION = '0.20';
 
 my %CLIENTS;
 my %AUTH_CODES;
 my %ACCESS_TOKENS;
 my %REFRESH_TOKENS;
+my $JWT_SECRET;
 
 =head1 METHODS
 
@@ -236,6 +253,8 @@ sub register {
   my $auth_route   = $config->{authorize_route}    // '/oauth/authorize';
   my $atoken_route = $config->{access_token_route} // '/oauth/access_token';
 
+  $JWT_SECRET = $config->{jwt_secret} // undef;
+
   if (
     # if we don't have a list of clients
     ! exists( $config->{clients} )
@@ -326,9 +345,8 @@ sub _authorization_request {
   if ( $res ) {
 
     $self->app->log->debug( "OAuth2::Server: Generating auth code for $c_id" );
-    my ( $auth_code,$expires_in ) = _generate_authorization_code(
-      $config->{auth_code_ttl}
-    );
+    my $expires_in = $config->{auth_code_ttl} // 600;
+    my $auth_code = _token( $expires_in,$c_id,\@scopes,'auth',$url );
 
     my $store_auth_code = $config->{store_auth_code} // \&_store_auth_code;
     $store_auth_code->( $self,$auth_code,$c_id,$expires_in,$url,@scopes );
@@ -379,16 +397,17 @@ sub _access_token_request {
 
   my $json_response = {};
   my $status        = 400;
-  my ( $client,$error,$scope,$old_refresh_token );
+  my ( $client,$error,$scope,$old_refresh_token,$user_id );
 
   if ( $grant_type eq 'refresh_token' ) {
-    ( $client,$error,$scope ) = _verify_access_token_and_scope(
+    ( $client,$error,$scope,$user_id ) = _verify_access_token_and_scope(
       $app,$config,$refresh_token,$self
     );
     $old_refresh_token = $refresh_token;
   } else {
-    my $verify_auth_code_sub = $config->{verify_auth_code} // \&_verify_auth_code;
-    ( $client,$error,$scope ) = $verify_auth_code_sub->(
+    my $verify_auth_code_sub = $config->{verify_auth_code}
+      // \&_verify_auth_code;
+    ( $client,$error,$scope,$user_id ) = $verify_auth_code_sub->(
       $self,$client_id,$client_secret,$auth_code,$url
     );
   }
@@ -397,8 +416,9 @@ sub _access_token_request {
 
     $self->app->log->debug( "OAuth2::Server: Generating access token for $client" );
 
-    my ( $access_token,$refresh_token,$expires_in )
-      = _generate_access_token( $config->{access_token_ttl} );
+    my $expires_in    = $config->{access_token_ttl} // 3600;
+    my $access_token  = _token( $expires_in,$client,$scope,'access',undef,$user_id );
+    my $refresh_token = _token( undef,$client,$scope,'refresh',undef,$user_id );
 
     my $store_access_token_sub
       = $config->{store_access_token} // \&_store_access_token;
@@ -435,36 +455,44 @@ sub _access_token_request {
   );
 }
 
-sub _generate_authorization_code {
-  my ( $ttl ) = @_;
+sub _token {
+  my ( $ttl,$client_id,$scopes,$type,$redirect_url,$user_id ) = @_;
 
-  $ttl //= 600;
-  my ( $sec,$usec ) = gettimeofday;
+  my $code;
 
-  return (
-    encode_base64( join( '-',$sec,$usec,rand(),random_string(30) ),'' ),
-    $ttl
-  );
-}
+  if ( ! $JWT_SECRET ) {
+    my ( $sec,$usec ) = gettimeofday;
+    $code = encode_base64( join( '-',$sec,$usec,rand(),random_string(30) ),'' );
+  } else {
+    $code = Mojo::JWT->new(
+      ( $ttl ? ( expires => time + $ttl ) : () ),
+      secret  => $JWT_SECRET,
+      set_iat => 1,
+      # https://tools.ietf.org/html/rfc7519#section-4
+      claims  => {
+        # Registered Claim Names
+#        iss    => undef, # us, the auth server / application (set using plugin config?)
+#        sub    => undef, # the logged in user, we could get this by returning it from the resource_owner_logged_in callback
+        aud    => $redirect_url, # the "audience"
+        jti    => random_string(32),
 
-sub _generate_access_token {
+        # Private Claim Names
+        user_id      => $user_id,
+        client       => $client_id,
+        type         => $type,
+        scopes       => $scopes,
+      },
+    )->encode;
+  }
 
-  my ( $ttl ) = @_;
-
-  $ttl //= 3600;
-
-  return (
-    ( _generate_authorization_code() )[0],
-    ( _generate_authorization_code() )[0],
-    $ttl,
-  );
+  return $code;
 }
 
 sub _verify_access_token_and_scope {
   my ( $app,$config,$refresh_token,$c,@scopes ) = @_;
 
-  my $verify_access_token_sub
-    = $config->{verify_access_token} // \&_verify_access_token;
+  my $verify_access_token_sub = $config->{verify_access_token}
+    // \&_verify_access_token;
 
   my $access_token;
 
@@ -671,6 +699,8 @@ the verify_auth_code callback for verification:
 =cut
 
 sub _store_auth_code {
+  return if $JWT_SECRET;
+
   my ( $c,$auth_code,$client_id,$expires_in,$uri,@scopes ) = @_;
 
   $AUTH_CODES{$auth_code} = {
@@ -692,12 +722,13 @@ Authorization Server. The callback is passed the Mojolicious controller object,
 the client_id, the client_secret, the authorization code, and the redirect uri.
 
 The callback should verify the authorization code using the rules defined in
-the reference RFC above, and return a list with 3 elements. The first element
+the reference RFC above, and return a list with 4 elements. The first element
 should be a client identifier (a scalar, or reference) in the case of a valid
 authorization code or 0 in the case of an invalid authorization code. The second
 element should be the error message in the case of an invalid authorization
 code. The third element should be a hash reference of scopes as requested by the
-client in the original call for an authorization code:
+client in the original call for an authorization code. The fourth element should
+be a user identifier:
 
   my $verify_auth_code_sub = sub {
     my ( $c,$client_id,$client_secret,$auth_code,$uri ) = @_;
@@ -740,12 +771,14 @@ client in the original call for an authorization code:
 
     $auth_codes->update( $ac,{ verified => 1 } );
 
-    return ( $client_id,undef,$scope );
+    return ( $client_id,undef,$scope,$ac->{user_id} );
   };
 
 =cut
 
 sub _verify_auth_code {
+  return _verify_auth_code_jwt( @_ ) if $JWT_SECRET;
+
   my ( $c,$client_id,$client_secret,$auth_code,$uri ) = @_;
 
   my ( $sec,$usec,$rand ) = split( '-',decode_base64( $auth_code ) );
@@ -794,6 +827,40 @@ sub _verify_auth_code {
 
 }
 
+sub _verify_auth_code_jwt {
+  my ( $c,$client_id,$client_secret,$auth_code,$uri ) = @_;
+
+  my $client = $CLIENTS{$client_id}
+    || return ( 0,'unauthorized_client' );
+
+  return ( 0,'invalid_grant' )
+    if ( $client_secret ne $client->{client_secret} );
+
+  my $auth_code_payload;
+
+  try {
+    $auth_code_payload = Mojo::JWT->new( secret => $JWT_SECRET )
+      ->decode( $auth_code );
+  } catch {
+    chomp;
+    $c->app->log->error( 'OAuth2::Server Auth code: ' . $_ );
+    return ( 0,'invalid_grant' );
+  };
+
+  if (
+    ! $auth_code_payload
+    or $auth_code_payload->{type} ne 'auth'
+    or $auth_code_payload->{client} ne $client_id
+    or ( $uri && $auth_code_payload->{aud} ne $uri )
+  ) {
+    return ( 0,'invalid_grant' );
+  }
+
+  my $scope = $auth_code_payload->{scopes};
+
+  return ( $client_id,undef,$scope,undef );
+}
+
 =head2 store_access_token
 
 A callback to allow you to store the generated access and refresh tokens. The
@@ -883,6 +950,8 @@ the verify_access_token callback for verification:
 =cut
 
 sub _store_access_token {
+  return if $JWT_SECRET;
+
   my (
     $c,$c_id,$auth_code,$access_token,$refresh_token,
     $expires_in,$scope,$old_refresh_token
@@ -1000,6 +1069,8 @@ message (almost certainly 'invalid_grant' in this case)
 =cut
 
 sub _verify_access_token {
+  return _verify_access_token_jwt( @_ ) if $JWT_SECRET;
+
   my ( $c,$access_token,$scopes_ref,$is_refresh_token ) = @_;
 
   if (
@@ -1049,6 +1120,46 @@ sub _verify_access_token {
   return ( 0,'invalid_grant' )
 }
 
+sub _verify_access_token_jwt {
+  my ( $c,$access_token,$scopes_ref,$is_refresh_token ) = @_;
+
+  my $access_token_payload;
+
+  try {
+    $access_token_payload = Mojo::JWT->new( secret => $JWT_SECRET )
+      ->decode( $access_token );
+  } catch {
+    chomp;
+    $c->app->log->error( 'OAuth2::Server Access token: ' . $_ );
+    return ( 0,'invalid_grant' );
+  };
+
+  if (
+    $access_token_payload
+    && (
+      $access_token_payload->{type} eq 'access'
+      || $is_refresh_token && $access_token_payload->{type} eq 'refresh'
+    )
+  ) {
+
+    if ( $scopes_ref ) {
+      foreach my $scope ( @{ $scopes_ref // [] } ) {
+        if ( ! grep { $_ eq $scope } @{ $access_token_payload->{scopes} } ) {
+          $c->app->log->debug(
+            "OAuth2::Server: Access token does not have scope ($scope)"
+          );
+          return ( 0,'invalid_grant' );
+        }
+      }
+    }
+
+    return $access_token_payload;
+  }
+
+  $c->app->log->debug( "OAuth2::Server: Access token does not exist" );
+  return 0;
+}
+
 1;
 
 =head1 PUTTING IT ALL TOGETHER
@@ -1079,19 +1190,52 @@ oauth also becomes available to call in various controllers, templates, etc:
 There are more examples included with this distribution in the examples/ dir.
 See examples/README for more information about these examples.
 
-=head1 CLIENT SECRET AND TOKEN SECURITY
+=head1 CLIENT SECRET, TOKEN SECURITY, AND JWT
 
-The auth codes and access tokens generated by the plugin should be unique. They
-use a combination of the generation time (to microsecond precision) + rand() +
-a call to Crypt::PRNG's random_string function. These are then base64 encoded to
-make sure there are no problems with URL encoding.
+The auth codes and access tokens generated by the plugin should be unique. When
+jwt_secret is B<not> supplied they are generated using a combination of the
+generation time (to microsecond precision) + rand() + a call to Crypt::PRNG's
+random_string function. These are then base64 encoded to make sure there are no
+problems with URL encoding.
+
+If jwt_secret is set, which should be a strong secret, the tokens are created
+with the L<Mojo::JWT> module and each token should contain a jti using a call
+to Crypt::PRNG's random_string function. You can decode the tokens, typically
+with L<Mojo::JWT>, to get the information about the client and scopes - but you
+should not trust the token unless the signature matches.
+
+As the JWT contains the client information and scopes you can, in theory, use
+this information to validate an auth code / access token / refresh token without
+doing a database lookup. However, it gets somewhat more complicated when you
+need to revoke tokens. For more information about JWTs and revoking tokens see
+L<https://auth0.com/blog/2015/03/10/blacklist-json-web-token-api-keys/> and
+L<https://tools.ietf.org/html/rfc7519>. Ultimately you're going to have to use
+some shared store to revoke tokens, but using the jwt_secret config setting means
+you can simplify parts of the process as the JWT will contain the client, user,
+and scope information (JWTs are also easy to debug: L<http://jwt.io>).
+
+When using JWTs expiry dates will be automatically checked (L<Mojo::JWT> has this
+built in to the decoding) and the hash returned from the call to ->oauth will
+look something like this:
+
+  {
+    'iat'     => 1435225100,               # generation time
+    'exp'     => 1435228700,               # expiry time
+    'aud'     => undef                     # redirect uri in case of type: auth
+    'jti'     => 'psclb1AcC2OjAKtVJRg1JjRJumkVTkDj', # unique
+
+    'type'    => 'access',                 # auth, access, or refresh
+    'scopes'  => [ 'list','of','scopes' ], # as requested by client
+    'client'  => 'some client id',         # as returned from verify_auth_code
+    'user_id' => 'some user id',           # as returned from verify_auth_code
+  };
 
 Since a call for an access token requires both the authorization code and the
 client secret you don't need to worry too much about protecting the authorization
 code - however you obviously need to make sure the client secret and resultant
 access tokens and refresh tokens are stored securely. Since if any of these are
 compromised you will have your app endpoints open to use by who or whatever has
-stolen them.
+access to them.
 
 You should therefore treat the client secret, access token, and refresh token as
 you would treat passwords - so hashed, salted, and probably encrypted. As with
@@ -1118,6 +1262,8 @@ L<https://stackoverflow.com/questions/4419915/how-to-keep-the-oauth-consumer-sec
 
 L<Mojolicious::Plugin::OAuth2> - A client side OAuth2 Mojolicious plugin
 
+L<Mojo::JWT> - encode/decode JWTs
+
 =head1 AUTHOR
 
 Lee Johnson - C<leejo@cpan.org>
